
#include <Python.h>
#include <stdlib.h>
#include <string.h>
#include <libsapilektor.h>


/* from libspeechd.h */
typedef enum{
    SPD_MALE1 = 1,
    SPD_MALE2 = 2,
    SPD_MALE3 = 3,
    SPD_FEMALE1 = 4,
    SPD_FEMALE2 = 5,
    SPD_FEMALE3 = 6,
    SPD_CHILD_MALE = 7,
    SPD_CHILD_FEMALE = 8
}SPDVoiceType;

struct int_constans {
    char *name;
    int value;
} _consts[]={
    {"SPD_MALE1", SPD_MALE1},
    {"SPD_MALE2", SPD_MALE2},
    {"SPD_MALE3", SPD_MALE3},
    {"SPD_FEMALE1", SPD_FEMALE1},
    {"SPD_FEMALE2", SPD_FEMALE2},
    {"SPD_FEMALE3", SPD_FEMALE3},
    {"SPD_CHILD_MALE", SPD_CHILD_MALE},
    {"SPD_CHILD_FEMALE", SPD_CHILD_FEMALE},
    
    {NULL,0} // sentinel
};

struct SapiLektor {
    char *vname;
    int sd_code; // sd_MALE1 etc
    struct SAPI_data *handle;
    char lc[4];
};

typedef struct {
    PyObject HEAD;
    
    char **voice_table;
    int used_voices_no;
    struct SapiLektor *lectors;
    int current_voice;
    double pause_factor;
    int voice_types;
} s4l_SapiLektorObject;

static void _free_sl(s4l_SapiLektorObject *self)
{
    int i;
    if (self->lectors) {
        for (i=0;i<self->used_voices_no;i++) {
            if (self->lectors[i].handle) {
                SAPI_close(self->lectors[i].handle);
            }
        }
        free(self->lectors);
        self->lectors=NULL;
    }
    if (self->voice_table) {
        free(self->voice_table);
        self->voice_table = NULL;
    }
}


static int _pasuje(char *str,PyObject *list)
{
    if (!list) return 1;
    return 1;
}

/*
 *__init__(self,voicenames=None)
 */

static int S4lObject_init(s4l_SapiLektorObject *self,
                        PyObject * args, PyObject *kwds)
{
    PyObject *languages=NULL;
    PyObject *voices=NULL;
    int i,n;
    char *err;
    struct SAPI_data *data;
    static char *kwlist[]={"languages","voices",NULL};

    if (! PyArg_ParseTupleAndKeywords(args, kwds, "|O!O!", kwlist,
		&PyList_Type,&languages,
		&PyList_Type,&voices)) return -1;
    self->voice_table=SAPI_voices();
    for (i=0;self->voice_table[i];i++);
    if (!i) {
	PyErr_SetString(PyExc_RuntimeError,"No voices found");
	return -1;
    }
    self->lectors=malloc(sizeof(*self->lectors) * i);
    self->used_voices_no=0;
    memset(self->lectors,0,sizeof(*self->lectors) * i);
    for (i=n=0;self->voice_table[i];i++) {
        if (!_pasuje(self->voice_table[i],voices)) continue;
        data=SAPI_open(self->voice_table[i],&err);
        if (!data) {
            PyErr_SetString(PyExc_RuntimeError,err);
            free(err);
            _free_sl(self);
            return -1;
        }
        strncpy(self->lectors[n].lc,data->lang,2);
        self->lectors[n].lc[2]=0;
        if (!_pasuje(self->lectors[n].lc,languages)) {
            SAPI_close(data);
            continue;
        }
        self->lectors[n].handle = data;
        self->lectors[n].vname=self->voice_table[i];
        self->used_voices_no=++n;
    }
    if (!self->used_voices_no) {
        _free_sl(self);
        PyErr_SetString(PyExc_RuntimeError,"No voices selected");
	return -1;
    }
    self->current_voice=0;
    self->pause_factor = 1.0;
    
    /* now we should set sd_code fields */
    
    self->voice_types = 0;
    for (i=0;i<self->used_voices_no;i++) {
        int mv,mxv;
        for (mv=mxv=0;mv<i;mv++) {
            if (strcmp(self->lectors[mv].lc,self->lectors[i].lc)) continue;
            if (self->lectors[mv].handle->sex != self->lectors[i].handle->sex) continue;
            mxv++;
        }
        if (self->lectors[i].handle->sex & 2) {
            if (mxv) self->lectors[i].sd_code = 0;
            else self->lectors[i].sd_code =
                SPD_CHILD_MALE + (self->lectors[i].handle->sex & 1);
        }
        else {
            if (mxv >= 3) self->lectors[i].sd_code = 0;
            else self->lectors[i].sd_code =
                SPD_MALE1 + mxv + 3* (self->lectors[i].handle->sex & 1);
        }
        self->voice_types |= 1 << self->lectors[i].sd_code;
    }
    
    /* all done */
    
    return 0;
}

#define SELF ((s4l_SapiLektorObject *)self)

static void S4lObject_dealloc(PyObject *self)
{
    _free_sl(SELF);
    self->ob_type->tp_free(self);
}


static PyObject *S4l_SetVoice(PyObject *self,PyObject *args)
{
    int vtype=0;
    int i;
    char *this_lang=SELF->lectors[SELF->current_voice].lc;
    if (!PyArg_ParseTuple(args,"i",&vtype)) {
	return NULL;
    }
    if (vtype < SPD_MALE1 || vtype > SPD_CHILD_FEMALE) {
        PyErr_SetString(PyExc_RuntimeError,"Voice not found");
    }
    for (i=0;i<SELF->used_voices_no;i++) {
        if (SELF->lectors[i].sd_code != vtype) continue;
        if (strcmp(SELF->lectors[i].lc,this_lang)) continue;
        SELF->current_voice=i;
        Py_RETURN_NONE;
    }
    PyErr_SetString(PyExc_RuntimeError,"Voice not found");
    return NULL;
}

static PyObject *S4l_SetSynthesisVoice(PyObject *self,PyObject *args)
{
    char *s;
    int i;
    if (!PyArg_ParseTuple(args,"s",&s)) {
	return NULL;
    }
    for (i=0;i<SELF->used_voices_no;i++) {
        if (strcasecmp(SELF->lectors[i].vname,s)) continue;
        SELF->current_voice=i;
        Py_RETURN_NONE;
    }
    PyErr_SetString(PyExc_RuntimeError,"Voice not found");
    return NULL;
}
    

static PyObject *S4l_SetLanguage(PyObject *self,PyObject *args)
{
    Py_RETURN_NONE;
}


static PyObject *S4l_SetPauseFactor(PyObject *self,PyObject *args)
{
    Py_RETURN_NONE;
}


static PyObject *S4l_GetWave(PyObject *self,PyObject *args)
{
    /*
     * returns tuple (wavedata, frequency)
     */
    
    char *text;
    int trimmed=1;
    short *wave;
    char *err;
    int offset,count,rest;
    PyObject *rc;
    
    if (!PyArg_ParseTuple(args,"si",&text,&trimmed)) {
	return NULL;
    }
    wave=SAPI_getSentence(SELF->lectors[SELF->current_voice].handle,
                text,&offset,&count,&rest,SELF->pause_factor,&err);
    if (!wave) {
        PyErr_SetString(PyExc_RuntimeError,err);
        free(err);
        return NULL;
    }
    if (!trimmed) {
        count += offset + rest;
        offset = 0;
        rest = 0;
    }
    rc=Py_BuildValue("(s#i)",
                     (char *)(wave + offset),
                     2*count,
                     SELF->lectors[SELF->current_voice].handle->freq);
    free(wave);
    return rc;
}

static PyObject *S4l_GetVoiceList(PyObject *self)
{
    PyObject *list;
    PyObject *tuple;
    int i;
    list=Py_BuildValue("[]");
    for (i=0;i<SELF->used_voices_no;i++) {
        tuple=Py_BuildValue("(ssssi)",
            SELF->lectors[i].vname,
            SELF->lectors[i].lc,
            (SELF->lectors[i].handle-> sex & 1)?"female":"male",
            (SELF->lectors[i].handle-> sex & 2)?"child":"adult",
            SELF->lectors[i].sd_code);
        PyList_Append(list,tuple);
    }
    return list;
}

static PyMethodDef S4lObject_methods[] = {
    {"SetVoice", (PyCFunction)S4l_SetVoice, METH_VARARGS,
     "SetVoice(name) - set current voice by name\n"},
    {"SetSynthesisVoice", (PyCFunction)S4l_SetSynthesisVoice, METH_VARARGS,
     "SetSynthesisVoice(type) - set current voice by type\n"},
    {"SetPauseFactor", (PyCFunction)S4l_SetPauseFactor, METH_VARARGS,
     "SetPauseFactor(factor) - set pause factor for shortener (0.2 .. 1.0)\n"},
    {"GetVoiceList", (PyCFunction)S4l_GetVoiceList, METH_VARARGS,
     "GetVoiceList() - returns list of tuples (name,language,gender,age,sd_code)\n"},
    {"GetWave", (PyCFunction)S4l_GetWave, METH_VARARGS,
     "GetWave(text[,trimmed]) - tuple (wave,frequency)\n"},
    {NULL, NULL, 0, NULL}           /* sentinel */
}; 

static PyTypeObject S4l_ObjectType = {
    PyObject_HEAD_INIT(NULL)
    0,                         /*ob_size*/
    "sapilektor.Lektor",             /*tp_name*/
    sizeof(s4l_SapiLektorObject), /*tp_basicsize*/
    0,                         /*tp_itemsize*/
    (destructor)S4lObject_dealloc,                         /*tp_dealloc*/
    0,                         /*tp_print*/
    0, /*tp_getattr*/
    0,                         /*tp_setattr*/
    0,                         /*tp_compare*/
    0,                         /*tp_repr*/
    0,                         /*tp_as_number*/
    0,                         /*tp_as_sequence*/
    0,                         /*tp_as_mapping*/
    0,                         /*tp_hash */
    0,                         /*tp_call*/
    0,                         /*tp_str*/
    0,                         /*tp_getattro*/
    0,                         /*tp_setattro*/
    0,                         /*tp_as_buffer*/
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,        /*tp_flags*/
    "Sapi4Linux interface\n",           /* tp_doc */
    0,		               /* tp_traverse */
    0,		               /* tp_clear */
    0,		               /* tp_richcompare */
    0,		               /* tp_weaklistoffset */
    0,		               /* tp_iter */
    0,		               /* tp_iternext */
    S4lObject_methods,         /* tp_methods */
    0,                         /* tp_members */
    0,                         /* tp_getset */
    0,                         /* tp_base */
    0,                         /* tp_dict */
    0,                         /* tp_descr_get */
    0,                         /* tp_descr_set */
    0,                         /* tp_dictoffset */
    (initproc)S4lObject_init,  /* tp_init */
};

#ifndef PyMODINIT_FUNC	/* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif
PyMODINIT_FUNC
initsapilektor(void) 
{
    PyObject* m;
    int i;

    S4l_ObjectType.tp_new = PyType_GenericNew;
    if (PyType_Ready(&S4l_ObjectType) < 0)
        return;

    m = Py_InitModule3("sapilektor", NULL,
                       "Sapi4Linux interface module");
    Py_INCREF(&S4l_ObjectType);
    PyModule_AddObject(m, "Lektor", (PyObject *)&S4l_ObjectType);
    for (i=0;_consts[i].name;i++) {
        PyModule_AddObject(m,_consts[i].name,Py_BuildValue("i",_consts[i].value));
    }
}
